/**
 * @file	cpx_interface.c
 *
 * @author	Francois Clad
 * @date	2012-07-30
 * @version	0.1
 */


#include "cpx_interface.h"

static void
	cpx_error (CPXENVptr, const char *, int),
	cpx_print_solution (CPXENVptr, CPXLPptr, char *);

#if CONFLICT == CONF_DELETE
static void cpx_resolve_conflict (CPXENVptr, CPXLPptr, int, int);
#endif



void cpx_error (CPXENVptr env, const char *msg, int cpx_errcode) {
	char cpx_errstr[CPXMESSAGEBUFSIZE];

	if (CPXgeterrorstring(env, cpx_errcode, cpx_errstr) == NULL)
		fprintf(stderr, "%s: Unknown error code %5d.\n", msg, cpx_errcode);
	else
		fprintf(stderr, "%s: %s\n", msg, cpx_errstr);

	exit(EXIT_FAILURE);
}



#if CONFLICT == CONF_DELETE
void cpx_resolve_conflict (CPXENVptr env, CPXLPptr lp, int confrows, int confcols) {
	int status, confstat,
		*rowind, *rowbdstat, numrows,
		*colind, *colbdstat, numcols;

	if ((rowind = (int *)calloc(confrows, sizeof(int))) == NULL)
		errExit("malloc");
	if ((rowbdstat = (int *)calloc(confrows, sizeof(int))) == NULL)
		errExit("malloc");
	if ((colind = (int *)calloc(confcols, sizeof(int))) == NULL)
		errExit("malloc");
	if ((colbdstat = (int *)calloc(confcols, sizeof(int))) == NULL)
		errExit("malloc");

	if ((status = CPXgetconflict(env, lp, &confstat,
					rowind, rowbdstat, &numrows, 
					colind, colbdstat, &numcols)) != 0)
		cpx_error(env, "Failed to get conflict information", status);

	if (numrows > 0)
		CPXdelrows(env, lp, rowind[0], rowind[0]);

	free(rowind);
	free(rowbdstat);
	free(colind);
	free(colbdstat);

	PRINT_INFO("Row %d deleted.\n", rowind[0]);
}
#endif


void cpx_print_solution (CPXENVptr env, CPXLPptr lp, char *filename) {
	int		solnstat, solnmethod, solntype;
	int		cur_numcols;
	int		status	= 0;
	double	objval;
	double	*x		= NULL;
	FILE	*file	= NULL;
	int		surplus, storespace;
	char	**nameptr = NULL,
			*namestore = NULL;


	if ( filename == NULL )
		file = stdout;
	else if ( (file = fopen(filename, "w")) == NULL )
		errExit("fopen");

	solnstat = CPXgetstat (env, lp);

	if			( solnstat == CPX_STAT_UNBOUNDED ) {
		fprintf (file,"Model is unbounded\n");
		return;
	} else if	( solnstat == CPX_STAT_INFEASIBLE ) {
		fprintf (file,"Model is infeasible\n");
		return;
	} else if	( solnstat == CPX_STAT_INForUNBD ) {
		fprintf (file,"Model is infeasible or unbounded\n");
		return;
	}

	if ( (status = CPXsolninfo (env, lp, &solnmethod, &solntype, NULL, NULL)) != 0 )
		cpx_error (env, "Failed to obtain solution info", status);
	fprintf (file,"Solution status %d, solution method %d\n", solnstat, solnmethod);

	if ( solntype == CPX_NO_SOLN )
		cpx_error(env, "Solution not available", status);

	if ( (status = CPXgetobjval (env, lp, &objval)) != 0 )
		cpx_error(env, "Failed to obtain objective value", status);
	fprintf (file,"Objective value %.10g.\n", objval);

	cur_numcols = CPXgetnumcols (env, lp);

	// Retrieve column names
	CPXgetcolname (env, lp, NULL, NULL, 0, &surplus, 0, cur_numcols - 1);
	storespace = - surplus;

	if ((namestore = (char *) malloc (storespace * sizeof(char))) == NULL)
		errExit("malloc");
	if ((nameptr = (char **) malloc (cur_numcols * sizeof(char *))) == NULL)
		errExit("malloc");

   if ((status = CPXgetcolname (env, lp, nameptr, namestore, storespace,
                           &surplus, 0, cur_numcols - 1)) != 0)
	   cpx_error(env, "Failed to get column names", status);

	// Retrieve solution vector
	if ( (x = (double *) malloc (cur_numcols * sizeof(double))) == NULL )
		errExit("malloc");

	if ( (status = CPXgetx (env, lp, x, 0, cur_numcols - 1)) != 0 )
		cpx_error(env, "Failed to obtain primal solution", status);


	// Write out the solution
	for (int i = 0; i < cur_numcols; i++)
		fprintf (file,"%s: Value = %2.10g\n", nameptr[i], x[i]);

	if ( filename != NULL )
		fclose(file);

	free(nameptr);
	free(namestore);
}


void cpx_solve (char *filename, char *outfile, char *resfile) {
	CPXENVptr	env = NULL;
	CPXLPptr	lp = NULL;
	CPXFILEptr	lfile = NULL;
	int			status = 0;

#if CONFLICT == CONF_DELETE
	int			confrows, confcols;
	int			numdel = 0;
#elif CONFLICT == CONF_RELAX
	int			numrows;
	double		*rhs;
#endif

	PRINT_INFO ("Solving...\n");
	
	// Initialize the CPLEX environment	
	if ((env = CPXopenCPLEX (&status)) == NULL)
		cpx_error(env, "Could not open CPLEX environment", status);

	// Open log file
	if ((lfile = CPXfopen ("cplex.log", "w")) == NULL)
		cpx_error (env, "Unable to open log file", status);

	// Set log file
	if ((status = CPXsetlogfile (env, lfile)) != 0)
		cpx_error (env, "Unable to set log file", status);

	// Create the problem
	if ((lp = CPXcreateprob(env, &status, PROBLEM_NAME)) == NULL)
		cpx_error(env, "Failed to create LP", status);

	// Load problem data from the given file
	if ((status = CPXreadcopyprob(env, lp, filename, "LP")) != 0)
		cpx_error(env, "Failed to load problem data", status);

	// Turn on output to the screen
	/* if ((status = CPXsetintparam(env, CPX_PARAM_SCRIND, CPX_ON)) != 0)
		cpx_error(env, "Failure to turn on screen indicator", status);*/
	

#if CONFLICT == CONF_DELETE
	// Check for confilcting contraints
	if ((status = CPXrefineconflict(env, lp, &confrows, &confcols)) != 0)
		cpx_error(env, "Failed to refine conflicts", status);
	
	while (CPXgetstat(env, lp) != CPX_STAT_CONFLICT_FEASIBLE) {
		printf ("test\n");
		cpx_resolve_conflict(env, lp, confrows, confcols);
		if ((status = CPXrefineconflict(env, lp, &confrows, &confcols)) != 0)
			cpx_error(env, "Failed to refine conflicts", status);
		numdel++;
	}

	PRINT_INFO("All conflicts resolved: %d constraints deleted.\n", numdel);

#elif CONFLICT == CONF_RELAX
	// Retrieve the number of rows (constraints) in the problem
	numrows = CPXgetnumrows (env, lp);

	// Set relaxation mode
	if ((status = CPXsetintparam(env, CPX_PARAM_FEASOPTMODE, CPX_FEASOPT_MIN_INF)) != 0)
		cpx_error(env, "Failure to set relaxation mode", status);

	// Set constraints relaxation preferences
	if ( (rhs = (double *) malloc (numrows * sizeof(double))) == NULL)
		errExit("malloc");
	for (int i = 0; i < numrows; i++)
		rhs[i] = 1;

	// Compute min-cost relaxation
	if ( (status = CPXfeasopt(env, lp, rhs, NULL, NULL, NULL)) != 0 )
		cpx_error(env, "Failed to compute a minimum-cost relaxation", status);

#endif


	// Write the presolved problem in a file
	if ( outfile != NULL ) {
		if ((status = CPXwriteprob(env, lp, outfile, "LP")) != 0)
			cpx_error(env, "Failed to write pre-solved problem", status);
	}


	// Optimize !
	if ((status = CPXmipopt (env, lp)) != 0)
		cpx_error(env, "Failed to optimize the MILP", status);

	cpx_print_solution(env, lp, resfile);


	// Free up the problem
	if ((status = CPXfreeprob(env, &lp)) != 0)
		cpx_error(env, "CPXfreeprob failed", status);

	// Free up the CPLEX environment
	if ((status = CPXcloseCPLEX(&env)) != 0)
		cpx_error(env, "Could not close CPLEX environment", status);
}


