/**
 * @file	formulation.c
 *
 * @author	Francois Clad
 * @date	2012-06-21
 * @version	0.1
 */

#include "formulation.h"

struct _lp_matrix {
	int		*rows;
	int		*cols;
	double	*res;
	uint	size;
	uint	capacity;
	uint	nb_rows;
};


static uint			lp_generate_columns		(glp_prob *prob, Topology *topo);
static lp_matrix	*lp_generate_rows		(glp_prob *prob, Topology *topo);
static void			lp_add_row				(glp_prob *, lp_matrix *, Topology *, uint, uintll *, Pathll *, bool);
static void			lp_set_path				(Topology *, lp_matrix *, uintll *, uint, uint, bool);

static lp_matrix	*lp_create_matrix		(void);
static void			lp_extend_matrix		(lp_matrix *);


lp_matrix *lp_create_matrix () {
	lp_matrix *mat = NULL;

	if ((mat = (lp_matrix *)malloc(sizeof(lp_matrix))) == NULL)
		errExit("malloc");

	mat->size = 0;
	mat->capacity = INITIAL_MATRIX_CAPACITY;
	mat->nb_rows = 0;

	if ((mat->rows = (int *)calloc(mat->capacity, sizeof(int))) == NULL)
		errExit("calloc");
	if ((mat->cols = (int *)calloc(mat->capacity, sizeof(int))) == NULL)
		errExit("calloc");
	if ((mat->res = (double *)calloc(mat->capacity, sizeof(double))) == NULL)
		errExit("calloc");

	return mat;
}

void lp_extend_matrix (lp_matrix *mat) {
	mat->capacity *= 2;

	if ((mat->rows = (int *)realloc(mat->rows, mat->capacity * sizeof(int))) == NULL)
		errExit("realloc");
	if ((mat->cols = (int *)realloc(mat->cols, mat->capacity * sizeof(int))) == NULL)
		errExit("realloc");
	if ((mat->res = (double *)realloc(mat->res, mat->capacity * sizeof(double))) == NULL)
		errExit("realloc");
}

void lp_free_matrix (lp_matrix *mat) {
	free(mat->rows);
	free(mat->cols);
	free(mat->res);

	free(mat);
}


glp_prob *lp_generate_problem (Topology *topo) {
	glp_prob *prob = NULL;
	lp_matrix *mat = NULL;
	uint counter = 0;

	if ((prob = glp_create_prob()) == NULL)
		errExit("glp_create_prob");

	glp_set_prob_name(prob, "Weight inference");
	glp_set_obj_dir(prob, GLP_MIN);

	PRINT_INFO("Generating columns...\t");
	counter = lp_generate_columns(prob, topo);
	PRINT_INFO_PLUS("%u variables created (%u directed edges).\n", counter, topo->nb_edges);

	PRINT_INFO("Generating rows...\t");
	mat = lp_generate_rows(prob, topo);
	glp_load_matrix(prob, mat->size, mat->rows, mat->cols, mat->res);
	PRINT_INFO_PLUS("%u constraints added.\n", mat->nb_rows);

	return prob;
}

void lp_free_problem (glp_prob *prob) {
	glp_delete_prob(prob);
	glp_free_env();
}

void lp_solve_problem (glp_prob *prob) {
	glp_simplex(prob, NULL);
	//glp_intopt(prob, NULL);

	fprintf(stdout, "Result: %g\n", glp_get_obj_val(prob));
}

void lp_print_result (glp_prob *prob, FILE *file, uint size) {
	for (uint i = 1; i < size; i++)
		fprintf(file, "%s: %g\n", glp_get_col_name(prob, i), glp_get_col_prim(prob, i));
}

uint lp_generate_columns (glp_prob *prob, Topology *topo) {
	char *label = NULL;
	uint count = 0;

	// Set number of columns: number of directed edges
	glp_add_cols(prob, topo->nb_edges);

	// For each directed edge
	for (uint i = 0; i < topo->size; i++) {
		for (Linkll *it = topo->nodes[i]->neighbors; it != NULL; it = it->next) {
			// Set link name: src_node -> dst_node
			if ((label = (char *)malloc((
								strlen(topo->nodes[i]->label)
								+ strlen(topo->nodes[it->dst]->label) + 5)
							* sizeof(char))) == NULL)
				errExit("malloc");
			label = strcpy(label, topo->nodes[i]->label);
			label = strncat(label, " -> ", 4);
			label = strncat(label, topo->nodes[it->dst]->label, strlen(topo->nodes[it->dst]->label));
			glp_set_col_name(prob, it->id + 1, label);

			// Set constraint: link weights must be integers >= 1
			glp_set_col_bnds(prob, it->id + 1, GLP_LO, 1.0, 0.0);
			glp_set_col_kind(prob, it->id + 1, GLP_IV);

			// Set column coefficient in objective function
			glp_set_obj_coef(prob, it->id + 1, 1);

			count++;
		}
	}
	return count;
}


lp_matrix *lp_generate_rows (glp_prob *prob, Topology *topo) {
	lp_matrix *mat = NULL;
	Pathll *cpl = NULL,
		   *apl = NULL;
	
	mat = lp_create_matrix();

	for (uint i = 0; i < topo->size; i++) {
		for (uint j = 0; j < topo->size; j++) {
			if (i == j)
				continue;

			if ((cpl = get_chosen_paths(topo->nodes[i], i, j)) == NULL)
				continue;
			apl = get_alt_paths(topo, cpl, i, j);

			lp_add_row(prob, mat, topo, i, cpl->path, cpl->next, TRUE);
			lp_add_row(prob, mat, topo, i, cpl->path, apl, FALSE);

			freePathll(cpl);
			freePathll(apl);
		}
	}
	return mat;
}

void lp_add_row (glp_prob *prob, lp_matrix *mat, Topology *topo, 
		uint base_node, uintll *path, Pathll *pl, bool ecmp) {
	uint base_index = 0;

	if (path == NULL || (base_node == path->value && path->next == NULL))
			return;

	for (Pathll *pit = pl; pit != NULL; pit = pit->next) {
		if (pit->path == NULL || (base_node == pit->path->value && pit->path->next == NULL))
			continue;
		mat->nb_rows++;
		base_index = mat->size;
		PRINT_DEBUG("Row %u:", mat->nb_rows);
		glp_add_rows(prob, 1);
		glp_set_row_bnds(prob, mat->nb_rows, ecmp ? GLP_FX : GLP_UP, 0.0, 0.0);

		// Initialize matrix
		for (uint i = 1; i <= topo->nb_edges; i++) {
			mat->size++;
			if (mat->size >= mat->capacity)
				lp_extend_matrix(mat);

			mat->rows[mat->size] = mat->nb_rows;
			mat->cols[mat->size] = i;
			mat->res[mat->size] = 0.0;
		}

		// Fill matrix
		lp_set_path(topo, mat, path, base_index, base_node, FALSE);
		lp_set_path(topo, mat, pit->path, base_index, base_node, TRUE);
		PRINT_DEBUG_PLUS("\n");
	}
}

void lp_set_path (Topology *topo, lp_matrix *mat, uintll *path, uint base_index, uint prev, bool reversed) {
	Linkll *l = NULL;

	if (path == NULL)
		return;

	if (prev != path->value) {
		for (Linkll *it = topo->nodes[prev]->neighbors; it != NULL && l == NULL; it = it->next)
			if (it->dst == path->value)
				l = it;

		if (l == NULL) {
			PRINT_WARNING ("Link not found: %s(%u) -> %s(%u)\n", topo->nodes[prev]->label, prev, topo->nodes[path->value]->label, path->value);
			return;
		}

		mat->res[base_index + l->id + 1] += reversed ? -1.0 : 1.0;
		PRINT_DEBUG_PLUS(" %c x_%u", reversed ? '-' : '+', l->id + 1);
	}

	lp_set_path(topo, mat, path->next, base_index, path->value, reversed);
}

