/**
 * @file	main.c
 *
 * @author	Francois Clad
 * @date	2012-06-21
 * @version	0.1
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>

#include "base.h"
#include "topology.h"
#include "formulation.h"
#include "cpx_interface.h"

#define NUM_ARGS				2
#define VERSION					"0.1"
#define DEFAULT_PROBLEM_NAME	"problem.lp"

struct options {
	char	*problem_in;
	char	*problem_out;
	char	*problem_pre;
	char	*problem_map;
	char	*output;
};

static void	usage				(char *);
static void	version				(char *);
static void	command_parser		(int, char **, struct options *);
static void	generate_problem	(char *, char *, char *, char *);

static void export_mapping (glp_prob *, char *);


int main (int argc, char **argv) {
	struct options opts;

	command_parser(argc, argv, &opts);


	// Generate problem and write it in a file

	if (opts.problem_in == NULL)
		generate_problem(argv[optind], argv[optind+1], opts.problem_map, 
				opts.problem_out == NULL ? DEFAULT_PROBLEM_NAME : opts.problem_out);

	//if (opts.problem_map)
	//	return EXIT_SUCCESS;

	// Solve the problem using CPLEX
	
	if (opts.problem_out == NULL)
		cpx_solve(opts.problem_in == NULL ? DEFAULT_PROBLEM_NAME : opts.problem_in,
				opts.problem_pre, opts.output);

	return EXIT_SUCCESS;
}

void export_mapping (glp_prob *prob, char *filename) {
	FILE *file = NULL;
	int numcols;

	numcols = glp_get_num_cols(prob);

	if ((file = fopen(filename, "w")) == NULL)
		errExit("Unable to open file");

	for (int i = 1; i <= numcols; i++)
		fprintf(file, "%d %s\n", i, glp_get_col_name(prob, i));

	fclose(file);
}


void generate_problem (char *graph, char *traces, char *map, char *problem) {
	hashtable *labels = NULL;
	Topology *topo = NULL;
	glp_prob *prob = NULL;


	// Create hash table
	if ((labels = create_table(4)) == NULL)
		errExit("create_table");

	// Create topology structure
	topo = newTopology(); 

	// Import graph
	PRINT_INFO("Reading graph... \t");
	importGraph(topo, labels, graph);
	PRINT_INFO_PLUS("%u nodes successfully loaded.\n", topo->size);

	// Import traces
	PRINT_INFO("Reading traces...\t");
	importTraces2(topo, labels, traces);
	PRINT_INFO_PLUS("%u paths successfully loaded.\n", topo->nb_paths);

	// Clear labels table (no longer usefull)
	destroy_table(labels);

	// Generate LP problem
	prob = lp_generate_problem(topo);

	// Write problem in a file
	glp_write_lp(prob, NULL, problem);

	// Write column mapping
	if (map != NULL)
		export_mapping(prob, map);

	// Free any remaining allocated memory
	lp_free_problem(prob);
	freeTopology(topo);
}


void usage (char *progname) {
	printf("Usage: %s [options...] topology traces\n", progname);

	printf("Options:\n" \
		"\t-s, --solve filename\tSolve problem contained in the given file\n" \
		"\t-w, --write filename\tWrite problem in the given file\n" \
		"\t-p, --presolve filename\tWrite pre-solved problem in the given file\n" \
		"\t-o, --output filename\tWrite optimization result in the given file\n" \
		"\t-h, --help\t\tDisplay this help and exit\n" \
		"\t-v, --version\t\tDisplay the program version and exit\n");

	exit(EXIT_SUCCESS);
}

void version (char *progname) {
	printf("%s, version " VERSION "\n", progname);
	printf("  Compiled on %s, %s\n", __DATE__, __TIME__);

#if defined DEBUG
	printf("  \033[0;33mDebug mode\033[0;m\n");
#elif defined RELEASE
	printf("  \033[0;32mRelease mode\033[0;m\n");
#else
	printf("  \033[0;31mInvalid build\033[0;m (use \'make BUILD=release\' or \'make BUILD=debug\')\n");
#endif // BUILD

	exit(EXIT_SUCCESS);
}

void command_parser (int argc, char **argv, struct options *opts) {
	int optId = 0;
	const char *optString = "w:s:p:o:hv?";
	const struct option longOpts[] = {
		{ "write", 1, NULL, 'w' },
		{ "solve", 1, NULL, 's' },
		{ "presolve", 1, NULL, 'p' },
		{ "map", 1, NULL, 'm' },
		{ "output", 1, NULL, 'o' },
		{ "help", 0, NULL, 'h' },
		{ "version", 0, NULL, 'v' },
		{ NULL, 0, NULL, 0 }};
	
	opts->problem_in = NULL;
	opts->problem_out = NULL;
	opts->problem_pre = NULL;
	opts->problem_map = NULL;
	opts->output = NULL;

	while ((optId = getopt_long(argc, argv, optString, longOpts, NULL)) != -1) {
		switch (optId) {
			case 's':
				opts->problem_in = optarg;
				break;
			case 'w':
				opts->problem_out = optarg;
				break;
			case 'p':
				opts->problem_pre = optarg;
				break;
			case 'm':
				opts->problem_map = optarg;
				break;
			case 'o':
				opts->output = optarg;
				break;
			case 'h':
				usage(argv[0]);
				break;
			case 'v':
				version(argv[0]);
				break;
			default:
				break;
		}
	}
	
	if (opts->problem_in == NULL && argc - optind < NUM_ARGS) {
		PRINT_ERROR("insufficient arguments. Type '%s --help' for usage.\n", argv[0]);
		exit(EXIT_FAILURE);
	}
}

