/**
 * @file	topology.c
 *
 * @author	Francois Clad
 * @date	2012-06-21
 * @version	0.1
 */

#include "topology.h"

#define INITIAL_MAX_NODES	128


static void		printRoutesRec	(Topology *, uint, uint, bool *, FILE *);

static Linkll	*linkll_new		(uint id, uint dst, Linkll *next);
static void		linkll_free		(Linkll *link);



Linkll *linkll_new (uint id, uint dst, Linkll *next) {
	Linkll *link = NULL;

	if ((link = (Linkll *)malloc(sizeof(Linkll))) == NULL)
		errExit("malloc");

	link->id = id;
	link->dst = dst;
	link->next = next;

	return link;
}

void linkll_free (Linkll *link) {
	if (link == NULL)
		return;

	linkll_free(link->next);
	free(link);
}

Topology *newTopology () {
	Topology	*topo = NULL;

	// Allocate topology structure
	if ((topo = (Topology *)malloc(sizeof(Topology))) == NULL)
		errExit("malloc");
	topo->size = 0;
	topo->max_size = INITIAL_MAX_NODES;
	if ((topo->nodes = (Node **)calloc(topo->max_size, sizeof(Node *))) == NULL)
		errExit("calloc");

	return topo;
}


void importGraph (Topology *topo, hashtable *table, char *filename) {
	FILE	*file = NULL;
	char	line[BUFFER_SIZE] = "";
	regex_t	regex;
	bool	valid = FALSE;
	uint	link_src = 0,
			link_dst = 0,
			link_id = 0;

	// Open data file
	if ((file = fopen(filename, "r")) == NULL)
		errExit("Unable to open data file");

	// Read graph file line by line
	while (fgets(line, BUFFER_SIZE, file) != NULL) {
		valid = FALSE;
		if (regcomp(&regex, "^\t[0-9]\\+ -- [0-9]\\+", 0) == 0) {
			valid = !regexec(&regex, line, 0, NULL, 0);
			regfree(&regex);
		}
		if (!valid)
			continue;

		// Retrieve link source and destination identifiers
		line[strlen(line)-2] = '\0';
		link_dst = get_hid(table, topo, strstr(line, " -- ") + 4);
		line[strstr(line," ")-line] = '\0';
		link_src = get_hid(table, topo, line + 1);

		// Add link in both directions
		topo->nodes[link_src]->neighbors = linkll_new(link_id++, link_dst, topo->nodes[link_src]->neighbors);
		topo->nodes[link_dst]->neighbors = linkll_new(link_id++, link_src, topo->nodes[link_dst]->neighbors);
	}
	topo->nb_edges = link_id;

	// Close data file
	fclose(file);
}


void printRoutes (Topology *topo, uint id, char *filename) {
	FILE *file = NULL;
	bool visited[topo->size];

	if ((file = fopen(filename, "w")) == NULL)
		errExit("Unable to open file");

	for (uint i = 0; i < topo->size; i++)
		visited[i] = FALSE;

	fprintf(file, "digraph \"P(%s)\" {\n", topo->nodes[id]->label);

	printRoutesRec(topo, id, id, visited, file);

	fprintf(file, "}\n");

	fclose(file);
}

void printRoutesRec (Topology *topo, uint root, uint id, bool *visited, FILE *file) {
	if (visited[id])
		return;

	visited[id] = TRUE;
	for (uintll *it = topo->nodes[root]->dests[id]; it != NULL; it = it->next) {
		fprintf(file, "\t%s -> %s;\n", topo->nodes[id]->label, topo->nodes[it->value]->label);
		printRoutesRec(topo, root, it->value, visited, file);
	}
}


void freeTopology (Topology *topo) {
	for (uint i = 0; i < topo->size; i++) {
		for (uint j = 0; j < topo->size; j++) 
			freeUintll(topo->nodes[i]->dests[j]);
		linkll_free(topo->nodes[i]->neighbors);
		free(topo->nodes[i]->dests);
		free(topo->nodes[i]->label);
		free(topo->nodes[i]);
	}
	free(topo->nodes);
	free(topo);
}


uint get_hid (hashtable *table, Topology *topo, char *key) {
	uint id = -1;
	char *new_key = NULL;

	// Check whether key is already in table
	if ((id = search_value(table, key)) != (uint)(-1))
		return id;

	// Insert value in hash table
	if ((new_key = (char *)malloc(strlen(key)*sizeof(char))) == NULL)
		errExit("malloc");
	strcpy(new_key, key);
	insert_value(table, new_key, topo->size); 
	id = topo->size++;

	// Extend nodes array if necessary
	if (id >= topo->max_size) {
		topo->max_size *= 2;
		if ((topo->nodes = (Node **)realloc(topo->nodes, topo->max_size * sizeof(Node *))) == NULL)
			errExit("realloc");
	}

	// Initialize node structure
	if ((topo->nodes[id] = (Node *)malloc(sizeof(Node))) == NULL)
		errExit("malloc");
	topo->nodes[id]->label = new_key;
	topo->nodes[id]->dests = NULL;
	topo->nodes[id]->neighbors = NULL;

	return id;
}

