/**
 * @file	traces.c
 *
 * @author	Francois Clad
 * @date	2012-07-27
 * @version	0.1
 */

#include "traces.h"


#define WHITE 0
#define GRAY 1
#define BLACK 2


/**
 * @brief	Create a new path list for a given path
 *
 * @param	path	Pointer to the path to be added
 * @param	next	Pointer to the next element in the list
 *
 * @return	Pointer to the newly created list element, or NULL if an error occur
 */
static Pathll	*newPathll		(uintll *, Pathll *);


/**
 * @brief	Merge two path lists in one
 *
 * @param	pl1		Pointer to the first path list
 * @param	pl2		Pointer to the second path list
 *
 * @return	Pointer to the merged list
 */
static Pathll	*mergePathll	(Pathll *, Pathll *);



static bool		checkLoop		(uintll **, unsigned int, unsigned int);
static bool		checkLoopRec	(uintll **, unsigned int, char *);
static uintll	*add_link		(hashtable *, Topology *, char *, uintll *);
static uintll	*add_node		(hashtable *, char *, uintll *);
static void		add_path_all	(Topology *, uintll *);
static void		add_path		(Node *, uintll *);
static uintll	*find_path		(uintll *, Pathll *);
static Pathll	*add_alt_paths	(Pathll *, Pathll *, Pathll *);
static Pathll	*concatPathll	(Pathll *, Pathll *);
static uintll	*concat_paths	(uintll *, uintll *);
static void		break_cycles	(Topology *);
static void		printPathll		(Topology *, Pathll *, char *, FILE *);


void importTraces (Topology *topo, hashtable *labels, char *filename) {
	FILE *file = NULL;
	char line[BUFFER_SIZE] = "",
		 *token = NULL,
		 *last_token = NULL;
	uintll *cur_path = NULL;
	bool	new_line = TRUE;

	// Initialize paths
	for (uint i = 0; i < topo->size; i++)
		if ((topo->nodes[i]->dests = (uintll **)calloc(topo->size, sizeof(uintll *))) == NULL)
			errExit("calloc");

	// Open data file
	if ((file = fopen(filename, "r")) == NULL)
		errExit("Unable to open data file");

	// Read data file line by line
	while (fgets(line, BUFFER_SIZE, file) != NULL) {
		new_line = (line[strlen(line)-1] == '\n');
		token = strtok(line, "|");

		if (cur_path != NULL) {
			// Reconstruct link
			char *tmp_token = NULL;
			if ((tmp_token = (char *)malloc((strlen(last_token) + strlen(token) + 5) * sizeof(char))) == NULL)
				errExit("malloc");
			strcpy(tmp_token, last_token);
			strcat(tmp_token, token);
			cur_path = add_link(labels, topo, tmp_token, cur_path);
			free(tmp_token);
		}

		while ((token = strtok(NULL, "|")) != NULL)
			cur_path = add_link(labels, topo, (last_token = token), cur_path);

		if (new_line) {
			add_path_all(topo, cur_path);
			cur_path = NULL;
		}
	}

	// Close data file
	fclose(file);

	// Ensure there is no cycle in traces
	break_cycles(topo);
}


void
importTraces2 (Topology *topo, hashtable *labels, char *filename)
{
	FILE *file = NULL;
	char line[BUFFER_SIZE] = "",
		 *path = NULL;
	uintll *cur_path = NULL;

	// Initialize paths
	for (uint i = 0; i < topo->size; i++)
		if ((topo->nodes[i]->dests = (uintll **)calloc(topo->size, sizeof(uintll *))) == NULL)
			errExit("calloc");

	// Open data file
	if ((file = fopen(filename, "r")) == NULL)
		errExit("Unable to open data file");

	// Read data file line by line
	while (fgets(line, BUFFER_SIZE, file) != NULL) {
		/* Retrieve the path separator position*/
		path = strstr (line, " | ");
		if (path == NULL)
			continue;

		/* Shift path pointer to the first identifier */
		path += 3;

		for (char *token = strtok (path, " "); token != NULL; token = strtok (NULL, " "))
			cur_path = add_node (labels, token, cur_path);

		add_path_all(topo, cur_path);
		cur_path = NULL;
	}

	// Close data file
	fclose(file);

	// Ensure there is no cycle in traces
	break_cycles(topo);
}


uintll *
add_node (hashtable *table, char *node, uintll *path)
{
	uint id = 0;

	// Remove a potential '\n' from the string
	if (node[strlen(node)-1] == '\n')
		node[strlen(node)-1] = '\0';

	// Retrieve node identifier
	id = search_value (table, node);
	if (id == (uint)(-1))
		return path;

	// Add node to current path
	return newUintll(id, path);
}


void printPaths (Topology *topo, Pathll *cpl, Pathll *apl, char *filename) {
	FILE *file = NULL;


	if ((file = fopen(filename, "w")) == NULL)
		errExit("Unable to open file");

	fprintf(file, "digraph \"Paths\" {\n");

	printPathll(topo, cpl, "green", file);
	printPathll(topo, apl, "red", file);


	fprintf(file, "}\n");

	fclose(file);
}

void printPathll (Topology *topo, Pathll *pl, char *color, FILE *file) {
	bool printed[topo->size * topo->size];
	for (uint i = 0; i < topo->size * topo->size; i++)
		printed[i] = FALSE;

	for (Pathll *pit = pl; pit != NULL; pit = pit->next)
		if (pit->path != NULL && pit->path->next != NULL)
			for (uintll *it = pit->path; it->next != NULL; it = it->next)
				if (!printed[it->value * topo->size + it->next->value]) {
					printed[it->value * topo->size + it->next->value] = TRUE;
					fprintf(file, "\t%s -> %s [color=%s];\n",
							topo->nodes[it->value]->label,
							topo->nodes[it->next->value]->label,
							color);
				}
}


uintll *add_link (hashtable *table, Topology *topo, char *link, uintll *path) {
	uint	link_src = 0,
			link_dst = 0;
	regex_t	regex;
	bool	valid = FALSE;

	// Remove a potential '\n' from the string
	if (link[strlen(link)-1] == '\n')
		link[strlen(link)-1] = '\0';

	// Check whether the link is valid (between two known routers)
	if (regcomp(&regex, "^ID[0-9]\\+ -> ID[0-9]\\+$", 0) == 0) {
		valid = !regexec(&regex, link, 0, NULL, 0);
		regfree(&regex);
	}
	if (!valid)
		return path;

	// Retrieve link source and destination identifiers
	if ((link_dst = search_value(table, strstr(link, " -> ") + 4)) == (uint)(-1))
		return path;
	link[strstr(link," ")-link] = '\0';
	if ((link_src = search_value(table, link)) == (uint)(-1) || link_src == link_dst)
		return path;

	// Check whether the link is continuous
	if (path != NULL && path->value != link_src) {
		add_path_all(topo, path);
		path = NULL;
	}

	if (path == NULL)
		path = newUintll(link_src, NULL);

	// Add link to current path
	return newUintll(link_dst, path);
}


void add_path_all (Topology *topo, uintll *path) {
	uintll *actual_path = NULL;

	if (path == NULL || path->next == NULL)
		return;

	actual_path = newUintll(path->value, NULL);
	for (uintll *it = path->next; it != NULL; it = it->next)
		add_path(topo->nodes[it->value], (actual_path = newUintll(it->value, actual_path)));

	freeUintll(actual_path);
	freeUintll(path);
	
	topo->nb_paths++;
}


void add_path (Node *node, uintll *path) {
	if (path == NULL)
		return;

	for (uintll *it = path; it->next != NULL; it = it->next)
		if (searchUintll(node->dests[it->next->value], it->value) == NULL)
			node->dests[it->value] = addUintll(it->next->value, node->dests[it->value]);
}


Pathll *get_alt_paths (Topology *topo, Pathll *cpl, uint src, uint dst) {
	Pathll *paths = NULL;

	for (uint i = 0; i < topo->size; i++)
		if (i != src && i != dst)
			paths = add_alt_paths(concatPathll(
						get_chosen_paths(topo->nodes[src], src, i),
						get_chosen_paths(topo->nodes[i], i, dst)), cpl, paths);

	return paths;
}

Pathll *concatPathll (Pathll *pl1, Pathll *pl2) {
	Pathll *result = NULL;

	for (Pathll *pit1 = pl1; pit1 != NULL; pit1 = pit1->next)
		for (Pathll *pit2 = pl2; pit2 != NULL; pit2 = pit2->next)
			result = newPathll(concat_paths(
						copyUintll(pit1->path),
						copyUintll(pit2->path)), result);

	freePathll(pl1);
	freePathll(pl2);

	return result;
}

Pathll *add_alt_paths(Pathll *pl, Pathll *cpl, Pathll *apl) {
	if (pl == NULL)
		return apl;

	for (Pathll *pit = pl; pit != NULL; pit = pit->next) {
		if (find_path(pit->path, cpl) == NULL && find_path(pit->path, apl) == NULL)
			apl = newPathll(pit->path, apl);
		else
			freeUintll(pit->path);
	}

	for (Pathll *pit = pl->next; pit != NULL; pit = pit->next) {
		free(pl);
		pl = pit;
	}
	free(pl);

	return apl;
}


uintll *find_path (uintll *path, Pathll *pl) {
	if (path == NULL)
		return NULL;

	for (Pathll *pit = pl; pit != NULL; pit = pit->next) {
		uintll *p1 = path;
		uintll *p2 = pit->path;
		while (p1->value == p2->value) {
			if (p1->next == NULL && p2->next == NULL)
				return pit->path;						// Path found !
			if (p1->next == NULL || p2->next == NULL)
				break;									// Paths have different lengths
			p1 = p1->next;
			p2 = p2->next;
		}
	}

	return NULL;
}


uintll *concat_paths (uintll *p1, uintll *p2) {
	uint	last = 0;
	uintll	*reversed = NULL,
			*it1 = NULL,
			*it2 = NULL;

	for (uintll *it = p1; it != NULL; it = it->next)
		reversed = newUintll(it->value, reversed);

	it1 = reversed;
	it2 = p2;
	while (it1 != NULL && it2 != NULL && it1->value == it2->value) {
		last = it1->value;
		it1 = it1->next;
		it2 = it2->next;
		free(reversed);
		free(p2);
		reversed = it1;
		p2 = it2;
	}
	freeUintll(reversed);

	for (it1 = p1; it1->value != last; it1 = it1->next);
	freeUintll(it1->next);
	it1->next = p2;

	return p1;
}


Pathll *get_chosen_paths (Node *node, uint id, uint dest) {
	Pathll *paths = NULL;

	if (id == dest)
		return newPathll(newUintll(dest, NULL), NULL);

	for (uintll *it = node->dests[id]; it != NULL; it = it->next)
		paths = mergePathll(paths, get_chosen_paths(node, it->value, dest));

	for (Pathll *pit = paths; pit != NULL; pit = pit->next)
		pit->path = newUintll(id, pit->path);

	return paths;
}


// Pathll basic functions

Pathll *newPathll (uintll *path, Pathll *next) {
	Pathll *pl = NULL;

	if (path == NULL || (pl = (Pathll *)malloc(sizeof(Pathll))) == NULL)
		return NULL;

	pl->path = path;
	pl->next = next;

	return pl;
}

void freePathll (Pathll *pl) {
	if (pl == NULL)
		return;

	freeUintll(pl->path);
	freePathll(pl->next);
	free(pl);
}

Pathll *mergePathll (Pathll *pl1, Pathll *pl2) {
	Pathll *last = pl1;

	if (last == NULL)
		return pl2;

	if (pl2 != NULL) {
		for (; last->next != NULL; last = last->next);
		last->next = pl2;
	}

	return pl1;
}

void break_cycles (Topology *topo) {
	for (uint i = 0; i < topo->size; i++)
		checkLoop(topo->nodes[i]->dests, topo->size, i);
}

bool checkLoop (uintll **graph, unsigned int size, unsigned int root) {
	char color[size];

	memset(color, WHITE, size * sizeof(char));
	
	return !checkLoopRec(graph, root, color);
}

bool checkLoopRec (uintll **graph, unsigned int node, char* color) {
	color[node] = GRAY;
	uintll *last = graph[node];
	for (uintll *it = graph[node]; it != NULL; it = it->next) {
		if (color[it->value] == GRAY) { // LOOP !
			if (it == graph[node])
				graph[node] = it->next;
			else
				last->next = it->next;
			free(it);
			return TRUE;
		} else if (color[it->value] == WHITE && checkLoopRec(graph, it->value, color))
			return TRUE;
	}
	color[node] = BLACK;
	return FALSE;
}

