#!/usr/bin/python2.7
# encoding: utf-8
'''
PreprocessTraces -- Merge tracelb result files and replace the IPs by a numeric id.

@author:     Simon Liénardy

@copyright:  2014 University of Liège. All rights reserved.

@contact:    simon.lienardy@ulg.ac.be
'''

import sys
import os
import re

import cPickle as pk
import networkx as nx

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter

__all__ = []
__version__ = 0.1
__date__ = '2014-09-09'
__updated__ = '2014-09-09'

DEBUG = 0
TESTRUN = 0
PROFILE = 0

# This will match IP and other things (like 300.987.1.0) but it is simpler
IP_REGEX = "(?:[0-9]{1,3}\.){3}[0-9]{1,3}"
TRACELB_REGEX = "((?:[0-9]{1,3}\.){3}[0-9]{1,3}) -> ((?:[0-9]{1,3}\.){3}[0-9]{1,3})"

def edges_to_paths(infile, outfile, ip_to_id, ip_source):
    """
    @param infile: path to file to process
    @param infile: path to result file
    @param ip_to_id: a {ip : id} dict wher ip is a string of an ip adress and id is an int
    @param ip_source: ip adress of the source
    """
    
    source_id = ip_to_id.get(ip_source, ip_source)
    
    begin_processed_trace = " ".join(('', '|', str(source_id), ''))
    
    if not hasattr(edges_to_paths, 'pattern'):
        edges_to_paths.pattern = re.compile(TRACELB_REGEX)
    
    # short translation function
    def translate_ip(ip_match):
        """
        @param ip_str: MatchObject of an IP
        @return: the corresponding id from the dict ip_to_id
        @requires: the dict ip_to_id 
        """
        # Take the string match, translate thank to the dict and convert the result into str.
        match_str1 = ip_match.group(1)
        match_str2 = ip_match.group(2)
        return ip_to_id.get(match_str1, match_str1), ip_to_id.get(match_str2, match_str2)
    
    edges_on_path = []
    
    for line in infile:
        
        if line.startswith("tracelb"):
            
            if not edges_on_path:
                continue
            
            dag = nx.DiGraph()
            dag.add_edges_from(edges_on_path)
            
            topo_sort = nx.topological_sort(dag)
            # Remind : index -1 is the end of the list
            paths = nx.all_simple_paths(dag, topo_sort[0], topo_sort[-1])
            
            for path in paths:
                outfile.write(begin_processed_trace + " ".join((str(i) for i in path)) + "\n")
                
            edges_on_path = []
        
        else:
            match = edges_to_paths.pattern.match(line)
            if match:
                edges_on_path += [translate_ip(match)]
    
    if edges_on_path:
        
        dag = nx.DiGraph()
        dag.add_edges_from(edges_on_path)
            
        topo_sort = nx.topological_sort(dag)
        # Remind : index -1 is the end of the list
        paths = nx.all_simple_paths(dag, topo_sort[0], topo_sort[-1])
            
        for path in paths:
            outfile.write(begin_processed_trace + " ".join((str(i) for i in path)) + "\n") 

class CLIError(Exception):
    '''Generic exception to raise and log different fatal errors.'''
    def __init__(self, msg):
        super(CLIError).__init__(type(self))
        self.msg = "E: %s" % msg
    def __str__(self):
        return self.msg
    def __unicode__(self):
        return self.msg

def main(argv=None): # IGNORE:C0111
    '''Command line options.'''

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
    program_shortdesc = __import__('__main__').__doc__.split("\n")[1]
    program_license = '''%s

  Created by Simon Liénardy on %s.
  Copyright 2014 University of Liège. All rights reserved.

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc, str(__date__))

    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument("-v", "--verbose", dest="verbose", action="count", help="set verbosity level [default: %(default)s]")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        parser.add_argument('-o', '--output', dest='output_file', help="name of the output file [default: %(default)s]", default="traces.out")
        parser.add_argument('mapping', help="pickle dump of a dict {k: v} = {IP: id}")
        parser.add_argument('traces', help="name of the input file", metavar="path", nargs='+')

        # Process arguments
        args = parser.parse_args()

        traces = args.traces
        verbose = args.verbose
        mapping_path = args.mapping
        output_file = args.output_file

        if verbose > 0:
            print("Verbose mode on")

        with open(mapping_path, 'r') as mapping_file:
            ip_to_id = pk.load(mapping_file)

        with open(output_file, 'w') as out_file:
            for  trace in traces:
                with open(trace, 'r') as in_file:
                    """
                    Temporaly broken : source ip is needed in te traces... (currently, it is the name of the file...)
                    """
                    edges_to_paths(in_file, out_file, ip_to_id)
                
        return 0
    
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 0


if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-h")
        sys.argv.append("-v")
        sys.argv.append("-r")
    if TESTRUN:
        import doctest
        doctest.testmod()
    if PROFILE:
        import cProfile
        import pstats
        profile_filename = 'PreprocessTraces_profile.txt'
        cProfile.run('main()', profile_filename)
        statsfile = open("profile_stats.txt", "wb")
        p = pstats.Stats(profile_filename, stream=statsfile)
        stats = p.strip_dirs().sort_stats('cumulative')
        stats.print_stats()
        statsfile.close()
        sys.exit(0)
    sys.exit(main())