#!/usr/bin/ python
# coding: utf-8
'''
Created on 26 juin 2014

@author: slienardy
'''

import argparse
import os
import sys
import networkx as nx
import cPickle as pk
import random

# First IP
IP_BEGINNING = 167837696 # = 10.1.0.0
IP_LIST_FILENAME = "iplist"
GV_FILENAME = "graph.gv"

class FakerouteFilesCreator(object):
    '''
    Generation of fakeroute input files
    '''


    def __init__(self, graph):
        '''
        Constructor
        '''
        self.graph = graph
        # Give an IP to all the nodes beginning at IP defined by IP_BEGINNING
        self.mapping = self.map_ip(graph.nodes_iter())
    
    def degree_one_nodes(self):
        return [node for (node, degree) in self.graph.to_undirected().degree_iter() if degree == 1]
    
    def generate_files(self, directory=".", inverse=False, weight_name="weight", template=None):
        
        if template is None:
            sources = self.degree_one_nodes()
            
            print directory
            
            self.__generate_files(sources, directory, inverse, weight_name)
        
    def __generate_files(self, sources, directory=".", inverse=False, weight_name="weight"):  
        """
        Generate files used by Fakeroute
        @param arg 
        @param src logical id of the source node in the graph
        @param 
        """

        graph = self.graph
      
        if inverse:
            graph = inverse_weight(graph, weight_name)
        
        mapping = self.mapping
        
        if not os.path.exists(directory):
            os.makedirs(directory)
            
        with open(directory + '/mapping.dump', 'w') as mapping_file:
            # Print on each line the node number followed by an attribued IP address
            #mapping_file.write("\n".join(("{} {}".format(k, v) for k,v in mapping.iteritems() )))
            # We should use cPickle possibilities to dump the IP mapping dict (reversed for back-translation):
            ip_to_id = {v : k for (k,v) in mapping.iteritems()}
            pk.dump(ip_to_id, mapping_file, protocol=0)
            
        with open(directory + '/mapping.txt', 'w') as mapping_file:
            mapping_file.writelines((str(k) + ' ' + v + '\n' for k,v in mapping.iteritems()))
                 
        for src in sources:
            
            src_ip = mapping[src]
            
            sub_directory = directory + "/" + src_ip
            
            if not os.path.exists(sub_directory):
                os.mkdir(sub_directory)
                
            # Take account of possible multiple shortest path :
            (predecessors, _ ) = nx.dijkstra_predecessor_and_distance(graph, src, weight=weight_name)

            autogen_nb = 1
            
            ip_list_file = open(sub_directory + "/" + IP_LIST_FILENAME, 'w')
            
            for k in graph.nodes_iter():
                if k == src:
                    continue
                
                if k in self.degree_one_nodes():
                    ip_list_file.write(mapping[k] + "\n")
                
                with open(sub_directory +"/"+mapping[k]+"-autogen"+str(autogen_nb), "w") as output_file:
                    
                    output_file.write("#DESC Automaticaly generated file no" + str(autogen_nb))
                    autogen_nb += 1
                    
                    node_stack = [k]
                    end = k
                    
                    while node_stack:
                        dest = node_stack.pop()
                        
                        src_list = predecessors[dest]
                        
                        d_str = "end" if (dest == end) else mapping[dest]
                        
                        for s in src_list:
                            if (s != src):
                                s_str = mapping[s]
                                node_stack.append(s)
                            else:
                                s_str = "start"
                            
                            output_file.write("\n" + s_str + " " + d_str)
                           
            ip_list_file.close()
        
        # Print a .gv (graphviz) graph    
        H = nx.Graph()
            
        for e in graph.edges():
            if e[0] != e[1]:
                H.add_edge(*e)
        
        nx.write_dot(nx.connected_component_subgraphs(H)[0], directory + '/' + GV_FILENAME)
        
    def print_probing_instructions(self, template, filepath):
        """
        Print a file that contains instructions for the probing
        """
        
        mapping = self.mapping
        nb_pot_sources = self.degree_one_nodes()
        
        
    def map_ip(self, id_iter):
        """
        Return a dictonary of the shape {id : ip}
        TODO: manage IP with the GML file.
        """
             
        ip = IPv4Adress(IP_BEGINNING)
        
        return  {i : str(ip.__iadd__(1)) for i in id_iter}
        
        


def inverse_weight(graph, weight_name):
    """
    Inverse the weight of a graph
    @param graph A networkx graph
    @param weight_name the string name of the weight among the attribute of the edge in the graph.
    """
    
    for edge in graph.edges_iter(data=True):
        edge[2][weight_name] = 1.0 / edge[2][weight_name]
    
    return graph


class MerlinParser(object):
    """
    Trivial class to parse merlin files
    """
    
    def get_and_parse(self, node, edge):
        
        strtotupl = str_to_tuple
        
        with open(node, "r") as node_file, open(edge, "r") as edge_file:
        
            edge_list = list(strtotupl(line) for line in edge_file.readlines())
            node_list = list(int(line) for line in node_file.readlines())
        
            return node_list, edge_list
        

def str_to_tuple(string):
    """
    Transform a string to tuple of numbers.
    @param str: a string of numbers (spaces distinguish the different numbers)
    @return: tuple of int, each int is a number from the string (order is conserved)
    @raise ValueError: if the string str does not contains string representation of numbers
    """
    
    # Split the string between spaces, transform each chunk in int and place them in a tuple.
    return tuple((lambda x : int(x))(chunk) for chunk in string.split())


class IPv4Adress():
    '''
    Fake IPV4 address handling
    '''
    MAX_IP = 2**32 - 1
    MIN_IP = 0

    def __init__(self, value=None):
        
        self._ip = value
        self.max_ip = IPv4Adress.MAX_IP
        self.min_ip = IPv4Adress.MIN_IP
    
    def __int__(self):
        return self._ip
        
    def __add__(self, other):
        """
        """
        
        if not isinstance(other, int):
            return NotImplemented

        if self._ip == self.max_ip and other > 0 :
            raise OverflowError("Max Ip recheaded")
        
        if self._ip == self.min_ip and other < 0 :
            raise OverflowError("Min Ip recheaded")

        return IPv4Adress(int(self) + other)
    
    def __iadd__(self, other):
        """
        """
        
        if not isinstance(other, int):
            return NotImplemented

        if self._ip == self.max_ip and other > 0 :
            raise OverflowError("Max Ip recheaded")
        
        if self._ip == self.min_ip and other < 0 :
            raise OverflowError("Min Ip recheaded")
        
        self._ip += other
        
        return self
        
    def __str__(self):
        
        masks = ((0xFF000000, 24), (0xFF0000, 16), (0xFF00, 8), (0xFF, 0))
        
        ip = self._ip
        
        octets = ((ip & mask) >> shift for (mask, shift) in masks)
        
        return ".".join(str(b) for b in octets)


def main(args):
    
    #Argparse module
    parser = argparse.ArgumentParser(description='EX2: --file path_to_gml')
    parser.add_argument('--file', metavar='F', default='', help='The path to the configuration file')
    parser.add_argument('--dir', metavar='D', type=str, default='lab', help='The path where the folder will be create')
    parser.add_argument('-i', '--inverse', action='store_try-ue', help='If present, the weights will be inverted to compute shortest path')
    parser.add_argument('-w', '--weight', type=str, help='Name of the weight attribute')

    parsed_args = parser.parse_args(args) 

    directory = parsed_args.dir
    configuration = parsed_args.file
    weight = parsed_args.weight
    inverse = parsed_args.inverse
        
    graph = nx.read_gml(configuration)
        
    f = FakerouteFilesCreator(graph)
    f.generate_files(directory, inverse, weight)
    
    return 0
    
if __name__ == '__main__':
    
    sys.exit(main(sys.argv))
    